LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;
USE ieee.std_logic_misc.all;
USE ieee.numeric_std.all;

-- *****************************************************************************
-- * This file define the behaviour of image processor.						   *
-- * The main functionalities of the image processor include converting RGB to *
-- * HSV, filter the pixels that pass the threshold comparison test and find the*
-- * center coordinate of the tracking object and stored it into a register.   *
-- * Created by Ryan Corpuz, Hang Peng										   *
-- *****************************************************************************

ENTITY RGB_to_HSV_converter IS 

-- *****************************************************************************
-- *                             Generic Declarations                          *
-- *****************************************************************************

-- *****************************************************************************
-- *                             Port Declarations                             *
-- *****************************************************************************
PORT (
		-- Inputs
		clk									:IN		STD_LOGIC;
		clk_en								:IN		STD_LOGIC;
		reset								:IN		STD_LOGIC;

		R									:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		G									:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		B									:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		
		H									:IN		STD_LOGIC_VECTOR( 8 DOWNTO  0);
		S									:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		V									:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		
		H_range								:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		S_range								:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		V_range								:IN		STD_LOGIC_VECTOR( 7 DOWNTO  0);
		
		
		thershould_range					:IN 	STD_LOGIC_vector(31 DOWNTO 0);
		stream_in_startofpacket				:IN		STD_LOGIC;

		-- Outputs
		location_x						:OUT STD_LOGIC_VECTOR(9 DOWNTO 0);
		location_y						:OUT STD_LOGIC_VECTOR(9 DOWNTO 0);
		
		temp1							:OUT STD_LOGIC_vector(31 DOWNTO 0);
		temp2							:OUT STD_LOGIC_vector(31 DOWNTO 0);
		temp3							:OUT STD_LOGIC_vector(31 DOWNTO 0);
		temp4							:OUT STD_LOGIC_vector(31 DOWNTO 0);
		
		sample_pixel_00					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_10					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_20					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_01					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_11					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_21					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_02					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_12					:OUT STD_LOGIC_vector(31 DOWNTO 0);
		sample_pixel_22					:OUT STD_LOGIC_vector(31 DOWNTO 0);
			
		stream_out_thresh				:BUFFER	STD_LOGIC

		
	);

END RGB_to_HSV_converter;

ARCHITECTURE Behaviour OF RGB_to_HSV_converter IS
-- *****************************************************************************
-- *                           Constant Declarations                           *
-- *****************************************************************************

-- *****************************************************************************
-- *                       Internal Signals Declarations                       *
-- *****************************************************************************
	
	-- Internal Wires
	SIGNAL RR :STD_LOGIC_VECTOR( 7 DOWNTO  0);
	SIGNAL GG :STD_LOGIC_VECTOR( 7 DOWNTO  0);
	SIGNAL BB :STD_LOGIC_VECTOR( 7 DOWNTO  0);
	
	SIGNAL 	MAXX								:INTEGER;
	SIGNAL 	MINN								:INTEGER;

	SIGNAL	R_norm							:INTEGER;	
	SIGNAL	G_norm							:INTEGER;
	SIGNAL	B_norm							:INTEGER;	
	
	SIGNAL  H_val								:INTEGER;
	SIGNAL	S_val								:INTEGER;	
	SIGNAL	V_val								:INTEGER;
	
	
	-- Internal Registers
	
	SIGNAL	startofpacket_shift_reg		:STD_LOGIC;	
	SIGNAL	endofpacket_shift_reg		:STD_LOGIC;	
	SIGNAL	valid_shift_reg				:STD_LOGIC;	
	
	-- Internal Wires	
	SIGNAL	first_indicator				:STD_LOGIC:= '0';	
	SIGNAL	x_start						:INTEGER := 0;
	SIGNAL	x_end						:INTEGER := 0;
	SIGNAL	y_start						:INTEGER := 0;
	SIGNAL	y_end						:INTEGER := 0;
	SIGNAL	x_index						:INTEGER := 0;
	SIGNAL	y_index						:INTEGER := 0;
	SIGNAL	counter						:INTEGER := 0;	
	SIGNAL	pass_fail					:STD_LOGIC;
	-- State Machine Registers
	
-- *****************************************************************************
-- *                          Component Declarations                           *
-- *****************************************************************************

-- max function will output the larger one of two integer numbers 
	
	function max (a,b : in integer) return integer is 
	begin
		if a > b then
			return a;
		else
			return b;
		end if;
	end max;
-- min function will output the smaller one of two integer numbers 	
	function min (a,b : in integer) return integer is 
	begin
		if a < b then
			return a;
		else
			return b;
		end if;
	end min;
	
	
BEGIN
-- *****************************************************************************
-- *                         Finite State Machine(s)                           *
-- *****************************************************************************


-- *****************************************************************************
-- *                             Sequential Logic                              *
-- *****************************************************************************

	
	PROCESS (clk)

	BEGIN
		IF clk'EVENT AND clk = '1' THEN
			IF (reset = '1') THEN
				H_val 	<= 0;
				S_val 	<= 0;
				V_val 	<= 0;
				R_norm	<= 0;
				G_norm	<= 0;
				B_norm	<= 0;
				x_start <= 160;
				x_end   <= 160;
				y_start <= 120;
				y_end	<= 120;
				x_index <= 0;
				y_index <= 0;
				counter <= 0;
				first_indicator <= '0';
			ELSIF (clk_en = '1') THEN
			-- converting the incoming RGB pixels to HSV format
				R_norm 		<= to_integer(unsigned(R));
				G_norm 		<= to_integer(unsigned(G));
				B_norm 		<= to_integer(unsigned(B));
				
				MAXX <= max(max(B_norm,R_norm),G_norm);
				MINN <= min(min(B_norm,R_norm),G_norm);
				
				V_val <= MAXX;
				S_val <= MAXX - MINN;
				
				IF S_val = 0 THEN
					H_val <= 0;
				ELSIF R_norm = MAXX THEN
					H_val <= (((G_norm-B_norm)*60)/S_val)  ;
					IF (H_val < 0) THEN
						H_val <= H_val + 360;
					END IF;
				ELSIF G_norm = MAXX THEN
					H_val <= (((B_norm - R_norm)*60)/S_val +120);
				ELSE
					H_val <= (((R_norm - G_norm)*60)/S_val +240);
				END IF;

--				-- Record the middle pixel of the frame
				IF (x_index = 160) AND (y_index = 120) THEN
					sample_pixel_11(14 DOWNTO 0) <= std_logic_vector(to_unsigned(H_val,15));
					sample_pixel_11(22 DOWNTO 15) <= std_logic_vector(to_unsigned(S_val,8));
					sample_pixel_11(30 DOWNTO 23) <= std_logic_vector(to_unsigned(V_val,8));
				END IF;

				-- Filter the converted HSV pixels by comparing each field to its pre-defined threshold range
				-- Also record the first pixel that pass the threshold comparison and keep updating the following pixel as last
				-- pixel for finding the cemter of the object,when it have all the pixels processed, it will output the coordinate
				-- by averaging x and y of the first and last pixels.
				IF (stream_in_startofpacket = '1') THEN
					IF (H_val < h_comp_max) AND (H_val > h_comp_min) AND (S_val < s_comp_max) AND (S_val > s_comp_min) AND (V_val < v_comp_max) AND (V_val > v_comp_min)THEN
						IF (first_indicator = '0') THEN
							y_start <= y_index;
							x_start <= x_index;
							first_indicator <= '1';
						ELSE
							x_end <= x_index;
							y_end <= y_index;
						END IF;
						pass_fail <= '1';
					ELSE
						pass_fail <= '0';
					END IF;
					IF (first_indicator = '1') AND (x_end = 160) AND (y_end = 120) THEN
						x_end <= x_start;
						y_end <= y_start;
					END IF;
					location_x(9 downto 0) <= std_logic_vector(to_unsigned(((x_end + x_start)/2),10));
					location_y(9 downto 0) <= std_logic_vector(to_unsigned(((y_end + y_start)/2),10));
					temp2 <= std_logic_vector(to_unsigned(counter,32));
					x_start <= 160;
					x_end   <= 160;
					y_start <= 120;
					y_end	  <= 120;
					x_index <= 0;
					y_index <= 0;
					first_indicator <= '0';
					counter <= 0;				
					x_index <= x_index + 1;
				ELSE
					IF (H_val < h_comp_max) AND (H_val > h_comp_min) AND (S_val < s_comp_max) AND (S_val > s_comp_min) AND (V_val < v_comp_max) AND (V_val > v_comp_min) THEN
						IF (first_indicator = '0') THEN
							y_start <= y_index;
							x_start <= x_index;
							first_indicator <= '1';
						ELSE
							x_end <= x_index;
							y_end <= y_index;
						END IF;
						pass_fail <= '1';
					ELSE
						pass_fail <= '0';
					END IF;				
					IF (x_index < 319) THEN
						x_index <= x_index + 1;
					ELSE
						x_index <= 0;
						y_index <= y_index +1;
					END IF;

						stream_out_thresh				<= pass_fail;
				END IF;
						counter <= counter + 1;
			END IF;
		END IF;
	END PROCESS;
	


END Behaviour;
